                .LF     PICNRC17.LST
;-----------------------------------------------------------------------------
;
; PIC-RC5.ASM
;
; PIC NRC-17 IR Receiver
;
; Program and hardware made by: San Bergmans Oisterwijk
;                               September 2004
;                               www.sbprojects.com
; Using the free SB-Assembler
;
;-----------------------------------------------------------------------------

                .CR     16CXXX          Use the PIC 16CXXX cross (16F84)
                .TF     PICNRC17.HEX,INT

;-----------------------------------------------------------------------------
;
; Special function file register definition
;
;-----------------------------------------------------------------------------

INDF            .EQ     $00             Indirect register
TMR0            .EQ     $01             Timer 0 value
PCL             .EQ     $02             Lower order 8 bits of PC
STATUS          .EQ     $03             Status flag register
FSR             .EQ     $04             File select register
PORTA           .EQ     $05             I/O port A
PORTB           .EQ     $06             I/O port B
EEDATA          .EQ     $08             EEPROM data register
EEADR           .EQ     $09             EEPROM address register
PCLATCH         .EQ     $0A             Upper 5 bits of PC
INTCON          .EQ     $0B             Interrupt control

OPTION          .EQ     $81             Option flag register
TRISA           .EQ     $85             Output direction register A
TRISB           .EQ     $86             Output direction register B
EECON1          .EQ     $88             EEPROM control register 1
EECON2          .EQ     $89             EEPROM control register 2

;-----------------------------------------------------------------------------
;
; Special function bit location definition
;
;-----------------------------------------------------------------------------

; STATUS register
CARRY           .EQ     0               Carry flag
DCARRY          .EQ     1               Decimal carry flag
ZERO            .EQ     2               Zero flag
PD              .EQ     3               Inverted power down bit
TO              .EQ     4               Inverted time out bit
RP0             .EQ     5               Register bank pointer 0
RP1             .EQ     6               Register bank pointer 1
IRP             .EQ     7               Indirect register bank pointer

; OPTION register
PS0             .EQ     0               Pre-scaler value 0
PS1             .EQ     1               Pre-scaler value 1
PS2             .EQ     2               Pre-scaler value 2
PSA             .EQ     3               Pre-scaler assign bit
T0SE            .EQ     4               Timer 0 source edge select bit
T0CS            .EQ     5               Timer 0 clock source select bit
INTEDG          .EQ     6               Interrupt edge select bit
RBPU            .EQ     7               Port B pull up enable bit

; INTCON register
RBIF            .EQ     0               RB port change interrupt flag bit
INTF            .EQ     1               External interrupt flag bit
T0IF            .EQ     2               Timer 0 overflow interrupt flag bit
RBIE            .EQ     3               RB port change interrupt enable bit
INTE            .EQ     4               INT interrupt enable bit
T0IE            .EQ     5               Timer 0 interrupt enable bit
EEIE            .EQ     6               EE write complete interrupt enable bit
GIE             .EQ     7               Global interrupt enable bit

;-----------------------------------------------------------------------------
;
; File register declarations
;
;-----------------------------------------------------------------------------

SCAN_STATE      .EQ     $0C             Scan state pointer
DIGIT1          .EQ     $0D             Pattern of left most digit
DIGIT2          .EQ     $0E
DIGIT3          .EQ     $0F
DIGIT4          .EQ     $10             Pattern of right most digit
SCAN_DELAY      .EQ     $11             Scan delay counter (5ms pace)
CLR_DELAY       .EQ     $12             Display clear delay counter
DP_DELAY        .EQ     $13             Decimal point delay counter
IR_STATE        .EQ     $14             IR receiver state machine
BIT_TIMER       .EQ     $15             Bit timer
IR_SHIFT        .EQ     $16,$17,$18,$19 IR shift register
FLAGS           .EQ     $1A             General purpose flag byte

;-----------------------------------------------------------------------------
;
; Constants
;
;-----------------------------------------------------------------------------

TIMER_UPDATE    .EQ     -50+3           Timer update value (2 dead cycles)
MS5_COUNT       .EQ     100             Number of counts to 5ms
CLR_TIME        .EQ     100             Display clear timeout (20ms/count)
DP_TIME         .EQ     3               Decimal point timeout (20ms/count)

HALF_TIME       .EQ     500/50          Number of timer counts per half bit

;-----------------------------------------------------------------------------
;
; Reset vector
;
;-----------------------------------------------------------------------------

RESET           GOTO    INIT            Get the system going

;-----------------------------------------------------------------------------
;
; Segment patterns
;
;-----------------------------------------------------------------------------

HEX2SEGMENTS    ANDLW   %0000.1111      Use only lower nibble
                ADDWF   PCL,F           Add offset to current PC

SEGMENTS        RETLW   %1100.0000      0
                RETLW   %1111.1001      1
                RETLW   %1010.0100      2
                RETLW   %1011.0000      3
                RETLW   %1001.1001      4
                RETLW   %1001.0010      5
                RETLW   %1000.0010      6
                RETLW   %1111.1000      7
                RETLW   %1000.0000      8
                RETLW   %1001.0000      9
                RETLW   %1000.1000      A
                RETLW   %1000.0011      B
                RETLW   %1100.0110      C
                RETLW   %1010.0001      D
                RETLW   %1000.0110      E
                RETLW   %1000.1110      F

;-----------------------------------------------------------------------------
;
; Scan display routine
;
;-----------------------------------------------------------------------------

SCAN_DISP       MOVLW   MS5_COUNT       Reload scan delay
                MOVWF   SCAN_DELAY
                MOVLW   %1111.1111      Switch off all Anode drivers

                MOVWF   PORTA            during the switch
                MOVF    SCAN_STATE,W    Jump to next display's routine
                MOVWF   PCL

;-------------------------------------------------------------SCAN DISPLAY 1--

SCAN_DISP1      MOVF    DIGIT1,W        Get pattern of left most digit
                MOVWF   PORTB            and send it out
                MOVLW   %1111.1110      Switch left most digit on
                MOVWF   PORTA
                MOVLW   #SCAN_DISP2     Next time do SCAN_DISP2
                MOVWF   SCAN_STATE
                RETURN

;-------------------------------------------------------------SCAN DISPLAY 2--

SCAN_DISP2      MOVF    DIGIT2,W        Get pattern of 2nd left digit
                MOVWF   PORTB            and send it out
                MOVLW   %1111.1101      Switch 2nd left digit on
                MOVWF   PORTA
                MOVLW   #SCAN_DISP3     Next time do SCAN_DISP3
                MOVWF   SCAN_STATE

                DECF    DP_DELAY,F      Decrement DP_DELAY every 20ms
                BTFSS   STATUS,ZERO
                RETURN                  We're done if counter <> 0 !

                BSF     DIGIT2,7        Clear Digit 2's decimal point now
                RETURN

;-------------------------------------------------------------SCAN DISPLAY 3--

SCAN_DISP3      MOVF    DIGIT3,W        Get pattern of 2nd right digit
                MOVWF   PORTB            and send it out
                MOVLW   %1111.1011      Switch 2nd right digit on
                MOVWF   PORTA
                MOVLW   #SCAN_DISP4     Next time do SCAN_DISP3
                MOVWF   SCAN_STATE
                RETURN

;-------------------------------------------------------------SCAN DISPLAY 4--

SCAN_DISP4      MOVF    DIGIT4,W        Get pattern of right most digit
                MOVWF   PORTB            and send it out
                MOVLW   %1111.0111      Switch right most digit on
                MOVWF   PORTA
                MOVLW   #SCAN_DISP1     Next time do SCAN_DISP1 again
                MOVWF   SCAN_STATE

                DECF    CLR_DELAY,F     Decrement CLR_DELAY every 20ms
                BTFSS   STATUS,ZERO
                RETURN                  We're done if counter <> 0 !

                MOVLW   %1011.1111      Only display 4 dashes
                MOVWF   DIGIT1
                MOVWF   DIGIT2
                MOVWF   DIGIT3
                MOVWF   DIGIT4
                BCF     FLAGS,7         Clear battery empty flag
                RETURN

;-----------------------------------------------------------------------------
;
; IR receiver state machine
;
;-----------------------------------------------------------------------------

IR_MACHINE      MOVF    IR_STATE,W      Jump to present state
                MOVWF   PCL

;---------------------------------------STATE 0, WAIT FOR BEGIN OF PRE-PULSE--

IR_STATE_0      BTFSC   PORTA,4         Input still high?
                RETURN                  Yes! Nothing to do

                MOVLW   HALF_TIME/2     Wait until we're in the center of the
                MOVWF   BIT_TIMER        pre-pulse
                MOVLW   IR_STATE_1      Next stop is state 1
                MOVWF   IR_STATE
                RETURN

;---------------------------STATE 1, PRE-PULSE DETECTED, CHECK IF IT IS REAL--

IR_STATE_1      DECFSZ  BIT_TIMER       Wait until center of pre-pulse
                RETURN                  Time's not up yet!

                BTFSC   PORTA,4         Is the input still low?
                GOTO    IR_ERROR_1      Nope! Exit with error

                MOVLW   HALF_TIME       Set interval to half a bit time
                MOVWF   BIT_TIMER
                MOVLW   5               Next 5 samples must be high!
                MOVWF   IR_SHIFT
                MOVLW   IR_STATE_2      Which is checked next
                MOVWF   IR_STATE
                RETURN

;-------------------STATE 2, PRE_PULSE DETECTED, NEXT 5 SAMPLES MUST BE HIGH--

IR_STATE_2      DECFSZ  BIT_TIMER       Wait until center of half bit
                RETURN                  Time's not up yet!

                BTFSS   PORTA,4         Is input high?
                GOTO    IR_ERROR_0      Nope! Exit with error

                MOVLW   HALF_TIME       Set interval to half a bit time again
                MOVWF   BIT_TIMER
                DECFSZ  IR_SHIFT        Decrement counter
                RETURN                  Counter <> 0, don't change state yet!

                MOVLW   IR_STATE_3      Go to next state
                MOVWF   IR_STATE
                CLRF    IR_SHIFT        Prepare shift register
                CLRF    IR_SHIFT+1
                RETURN

;---------STATE 3, EXPECTING START BIT, OR ANOTHER PAUSE IF BATTERY IS EMPTY--

IR_STATE_3      DECFSZ  BIT_TIMER       Wait until center of first half of
                RETURN                   start bit.

                BTFSS   PORTA,4         Is it the start bit (if input is low)?
                GOTO    IR_START_BIT    Yes! Handle it!

                BSF     FLAGS,7         Set battery empty flag
                MOVLW   HALF_TIME*2     Next stop is really the start bit!
                MOVWF   BIT_TIMER
                MOVLW   IR_STATE_4      Next time handle first half of a bit
                MOVWF   IR_STATE
                RETURN

;-----------------------------------IR STATE 4, WAIT FOR FIRST HALF OF A BIT--

IR_STATE_4      DECFSZ  BIT_TIMER       Wait until center of first half of bit
                RETURN                  Keep waiting!

                MOVLW   IR_STATE_5      Next state is 5 if input is low
                BTFSS   PORTA,4
IR_START_BIT    MOVLW   IR_STATE_6      Input is high, next state is 6
                MOVWF   IR_STATE

                MOVLW   HALF_TIME       Restart bit timer
                MOVWF   BIT_TIMER
                RETURN

;---------------IR STATE 5, FIRST HALF WAS HIGH NOW IT MUST BE LOW FOR A "0"--

IR_STATE_5      DECFSZ  BIT_TIMER       Wait until center of 2nd half of bit
                RETURN                  Keep waiting!

                BTFSC   PORTA,4         Is input high now?
                GOTO    .ERROR          Nope! It's an error!

                BCF     STATUS,CARRY    A 0 was received, shift it in result
                RRF     IR_SHIFT+1,F
                RRF     IR_SHIFT,F
                MOVLW   HALF_TIME       Restart bit timer
                MOVWF   BIT_TIMER
                MOVLW   IR_STATE_4      In case we need some more bits
                BTFSC   STATUS,CARRY    We're done when Carry is 1
                MOVLW   IR_STATE_7      Carry is 1, received entire message
                MOVWF   IR_STATE
                RETURN

.ERROR          MOVLW   IR_ERROR_0      Wait until input gets high before
                MOVWF   IR_STATE         returning to state 0
                RETURN

;---------------IR STATE 6, FIRST HALF WAS LOW NOW IT MUST BE HIGH FOR A "1"--

IR_STATE_6      DECFSZ  BIT_TIMER       Wait until center of 2nd half of bit
                RETURN                  Keep waiting!

                BTFSS   PORTA,4         Is input high now?
                GOTO    IR_ERROR_1      Nope! It's an error!

                BSF     STATUS,CARRY    A 1 was received, shift it in result
                RRF     IR_SHIFT+1,F
                RRF     IR_SHIFT,F
                MOVLW   HALF_TIME       Restart bit timer
                MOVWF   BIT_TIMER
                MOVLW   IR_STATE_4      In case we need some more bits
                BTFSC   STATUS,CARRY    We're done when Carry is 1
                MOVLW   IR_STATE_7      Carry is 1, message complete!
                MOVWF   IR_STATE
                RETURN

;--------------------------IR STATE 7, MESSAGE RECEIVED, START PROCESSING IT--

IR_STATE_7      INCF    IR_SHIFT+1,W    Could it be address $FF, command $FE?
                BTFSS   STATUS,ZERO
                GOTO    .SKIP           No, it can't be!
                MOVLW   2
                ADDWF   IR_SHIFT,W
                BTFSC   STATUS,ZERO
                GOTO    .QUIT           It is $FF, $FE. Don't display it

.SKIP           MOVLW   CLR_TIME        Set display clear timer
                MOVWF   CLR_DELAY
                MOVLW   DP_TIME         Flash receive LED
                MOVWF   DP_DELAY

                SWAPF   IR_SHIFT+1,W    Work from left to right
                CALL    HEX2SEGMENTS
                MOVWF   DIGIT1

                MOVF    IR_SHIFT+1,W    Do the same with 2nd digit
                CALL    HEX2SEGMENTS
                ANDLW   %0111.1111      Flash dot of this digit
                MOVWF   DIGIT2

                SWAPF   IR_SHIFT,W      And with the 3d digit
                CALL    HEX2SEGMENTS
                MOVWF   DIGIT3

                MOVF    IR_SHIFT,W      And finally with the last digit
                CALL    HEX2SEGMENTS
                MOVWF   DIGIT4
                BTFSC   FLAGS,7         Is battery empty?
                BCF     DIGIT4,7        Yes! Turn right most DP on

.QUIT           MOVLW   IR_STATE_8      Done enough for now. Let's finish it
                MOVWF   IR_STATE         in the last state
                RETURN

;----------------------------------IR STATE 8, WAIT FOR INPUT TO RETURN HIGH--

IR_STATE_8
IR_ERROR_0      MOVLW   IR_STATE_0      Reset state machine only if input is
                BTFSC   PORTA,4          high
                MOVWF   IR_STATE
                RETURN

;-----------------------------------------------------------IR ERROR STATE 1--

IR_ERROR_1      MOVLW   IR_STATE_0      Return to IR state 0
                MOVWF   IR_STATE
                RETURN

;-----------------------------------------------------------------------------
;
; Get started
;
;-----------------------------------------------------------------------------

INIT            MOVLW   #$0D            Clear memory
                MOVWF   FSR
                MOVLW   #35             Clear a total of 36 RAM addresses
                MOVWF   $0C
.CLRMEM         CLRF    INDF            Clear this address
                INCF    FSR             Increment pointer
                DECFSZ  $0C             Decrement counter
                GOTO    .CLRMEM         Repeat until done

                BSF     STATUS,RP0      Select bank 1 of register file
                MOVLW   #%1101.1111     Timer mode without pre-scaler
                MOVWF   OPTION
                MOVLW   #%0001.0000     Only PA4 is an input
                MOVWF   TRISA
                MOVLW   #%0000.0000     All Port B are outputs
                MOVWF   TRISB
                BCF     STATUS,RP0      Select bank 0 of register file

                MOVLW   #%0000.0000     Disable all interrupts
                MOVWF   INTCON

                MOVLW   IR_STATE_0      Init IR receiver state machine
                MOVWF   IR_STATE

                MOVLW   #SCAN_DISP1     Init display scan state
                MOVWF   SCAN_STATE

                CLRF    PCLATCH         All indirect jumps are on page 0!

                MOVLW   TIMER_UPDATE    Get the timer started
                MOVWF   TMR0
                MOVLW   MS5_COUNT       Start 5ms counter for display scan
                MOVWF   SCAN_DELAY

                MOVLW   %1010.1011      Show nrc on display
                MOVWF   DIGIT1
                MOVLW   %1010.1111
                MOVWF   DIGIT2
                MOVLW   %1010.0111
                MOVWF   DIGIT3
                MOVLW   %1111.1111
                MOVWF   DIGIT4
                MOVLW   200             Set clear time-out to 4 seconds
                MOVWF   CLR_DELAY

;-----------------------------------------------------------------------------
;
; Main program loop
;
;-----------------------------------------------------------------------------

MAIN

;-----------------------------------------CALL THE IR RECEIVER STATE MACHINE--

                CALL    IR_MACHINE      Call the IR receiver state machine

;-----------------------------------------------------SCAN DISPLAY EVERY 5MS--

                DECF    SCAN_DELAY,F    Decrement scan delay counter
                BTFSC   STATUS,ZERO     Only scan display if counter = 0
                CALL    SCAN_DISP

;-------------------------------------------------------SYNC MAIN WITH TIMER--

.SYNC           BTFSC   TMR0,7          Wait until bit 7 of TMR = 0
                GOTO    .SYNC           Not 0 yet!
                MOVLW   TIMER_UPDATE    Reload timer again
                ADDWF   TMR0,F

                GOTO    MAIN

;-----------------------------------------------------------------------------
;
; Configuration word
;
;-----------------------------------------------------------------------------

                .OR     $2007
                .CW     XT
;               .LI     OFF
